/**
*
* This program implements differential cryptanalysis to attack a 2-round modified DES.
*
* First task is to finish Feistel network, using the given P,K, test ciphertext. 
* Run encryption for 1000 times on school computers, no big variants, 2783603 ns, 2722802 ns, 2674251 ns, 2630483 ns.
* Safe to say, for each encryption, it takes about 2700000 ns.
* Thus on school computer, on average, 2^56/2 times encryption are needed, each 1000 encryptions take 2700000 ns
* Total time to break the cipher is about: 2^55*2700ns=9.73e+19 ns = 9,73e+10s = about 3085 years
*
* 
**/

import java.util.Random;
import java.io.*;
import java.math.BigInteger;
import java.util.HashSet;
import java.util.Iterator;
import java.util.*;

class DESPrac
{
	//Variables universally used
	static long[] inputDifferentials = new long[8];				//differentials for 8 subkeys
	static long[] outputDifferentials = new long[8];			//expected differentials for outputs
	static byte[] sboxinputDifferentials = new byte[8];			//differentials of sbox inputs, each 6-bit integer
	static byte[] sboxoutputDifferentials = new byte[8];		//differentials of sbox outputs, each 6-bit integer
	static int[] possibilityTable = new int[8];

    public static void main(String[] args) throws IOException
    {
    	//Inits

    	inputDifferentials[0]=0x0080820060000000L;
 		outputDifferentials[0]=0x0000000060000000L;
 		sboxinputDifferentials[0]=0x0c;
 		sboxoutputDifferentials[0]=0x0e;
 		possibilityTable[0]=14;

 		inputDifferentials[1]=0x4008000004000000L;
 		outputDifferentials[1]=0x0000000004000000L;
 		sboxinputDifferentials[1]=0x08;
 		sboxoutputDifferentials[1]=0x0a;
 		possibilityTable[1]=16;

 	//	inputDifferentials[2]=0x--------00800000L;
 		outputDifferentials[2]=0x0000000001000000L;
 		sboxinputDifferentials[2]=0x20;
 		sboxoutputDifferentials[2]=0x0d;
 		possibilityTable[2]=16;

 		//Hack the whole thing

 		printEBoxOutput(outputDifferentials[2], 3);

    	// Task 1

/*		long testP=0x1234567887654321L;
		long testK=0x33333333333333L;
		// ...
		long test2rounds=TwoRoundModifiedDES(testK,testP);
		System.out.println(String.format("%x",test2rounds));

		long time1000round_start;
		long time1000round_end;
		time1000round_start=System.nanoTime();
		for(int i=0;i<1000;i++)
		{
			TwoRoundModifiedDES(testK,testP);
		}
		time1000round_end=System.nanoTime();
		System.out.println(String.format("Start: %d\nend: %d\ntime passed: %d",time1000round_start,time1000round_end,time1000round_end-time1000round_start));
    	*/

		//Task 2

    	//differentialDistributionTable(3);

 		//Task 3

 /*		iosolateK(1);*/

 		//Task 4

// 		differentialDistributionTable(2);

/* 		long deltaR0=0x0000000004000000L;
 		long deltaE=EBox(deltaR0);
 		for(int i=0;i<8;i++)
 		{
 			System.out.println(String.format("Delta E is: %02x",(deltaE>>(42-6*i))&0x3f));
 		}
 		
 		long deltaO=0x000000000a000000L;
 		long deltaP=PBox(deltaO);
 		System.out.println(String.format("Delta P and Delta F is: %x",deltaP));

 		System.out.println("Isolate Sa:");

 		iosolateK(1);

 		System.out.println("Isolate Sb:");

 		iosolateK(2);*/

    }

    static long printEBoxOutput (long inputDifferentialR, int indexK)
    {
    	long deltaE=EBox(inputDifferentialR);
    	for(int i=0;i<8;i++)
 		{
 			System.out.println(String.format("Delta E is: %02x",(deltaE>>(42-6*i))&0x3f));
 		}
 		long nonZeroOut_deltaSK=(deltaE>>(42-6*(indexK-1))&0x3f);
 		return nonZeroOut_deltaSK;
    }

    static void findOutDifferential (long inputDifferentialR, int indexK, long deltaO)
    {
    	//TODO: Given input differential R0, produce output differential of first Feistal Net, so that left part of out is determined.
    	long deltaP=PBox(deltaO);
    	System.out.println(String.format("Delta P and Delta F is: %x",deltaP));
    }

    static void differentialDistributionTable (int SBoxNumber)
    {
    	if(SBoxNumber>8 || SBoxNumber<1)
    	{
    		System.out.println("SBoxNumber not in scope, program quits.");
    		System.exit(-1);
    	}
    	//calculate differential distribution table
    	int[][] countlist=new int[64][16];
    	for(int i=0;i<64;i++)							//i is delta S
    	{
    		for(int j=0;j<64;j++)						//j is possible numbers which with another number xor to delta O
    		{
    			int deltanumber=j^i;					//Given j, and know delta S is i, compute the other number
    			int deltaO=(STables[SBoxNumber-1][deltanumber]^STables[SBoxNumber-1][j]);
    			countlist[i][deltaO]++;
    		}
    	}
		//Display differential distribution table
		System.out.println("   |  0  1  2  3  4  5  6  7  8  9  A  B  C  D  E  F  ");
		System.out.println("---+--------------------------------------------------");
		for(int i=0;i<64;i++)
		{
			System.out.print(String.format("%02x |  ",i));
			for(int j=0;j<16;j++)
			{
				System.out.print(String.format("%02d ",countlist[i][j]));
			}
			System.out.println();
		}
    }

    // constants for &-ing with, to mask off everything but the bottom 32- or 48-bits of a long
    static long MASK32 = 0xffffffffL;
    static long MASK48 = 0xffffffffffffL;

    static long TwoRoundModifiedDES(long K, long P) // input is a 56-bit key "long" and a 64-bit plaintext "long", returns the ciphertext
    {
		long L0=(P>>32)&MASK32; // watch out for the sign extension! 
		long R0=P&MASK32;
		long K1=(K>>8)&MASK48;
		long K2=K&MASK48;

		long L1=R0;
		long R1=L0^Feistel(R0, K1);

		long L2=R1;
		long R2=L1^Feistel(R1, K2);

		long C=L2<<32 | R2;

		return(C);
    }

    static long Feistel(long R, long K) // input is a 32-bit integer and 48-bit key, both stored in 64-bit signed "long"s; returns the output of the Feistel round
    {
		long F;
		// ...
		long e1,k1_bottom,s1,sbox_output,p1;	//all intermediate variables used in a Feistel network
		k1_bottom=K&0xffffffffL;
		e1=EBox(R);
		s1=e1^K;
		sbox_output=SBox(s1);
		p1=PBox(sbox_output);
		F=p1^k1_bottom;
		return(F);
    }

    // NB: thecallEncryptse d HashSet <String>iffer from the tables in the DES standard because the latter are encoded in a strange order
    static final byte[] S1Table={
	14,  0,  4, 15, 13,  7,  1,  4,  2, 14, 15,  2, 11, 13,  8,  1,
	 3, 10, 10,  6,  6, 12, 12, 11,  5,  9,  9,  5,  0,  3,  7,  8,
	 4, 15,  1, 12, 14,  8,  8,  2, 13,  4,  6,  9,  2,  1, 11,  7,
	15,  5, 12, 11,  9,  3,  7, 14,  3, 10, 10,  0,  5,  6,  0, 13,
    };
    static final byte[] S2Table={
	15,  3,  1, 13,  8,  4, 14,  7,  6, 15, 11,  2,  3,  8,  4, 14,
	 9, 12,  7,  0,  2,  1, 13, 10, 12,  6,  0,  9,  5, 11, 10,  5,
	 0, 13, 14,  8,  7, 10, 11,  1, 10,  3,  4, 15, 13,  4,  1,  2,
	 5, 11,  8,  6, 12,  7,  6, 12,  9,  0,  3,  5,  2, 14, 15,  9,
    };

    static final byte[] S3Table={
	10, 13,  0,  7,  9,  0, 14,  9,  6,  3,  3,  4, 15,  6,  5, 10,
	 1,  2, 13,  8, 12,  5,  7, 14, 11, 12,  4, 11,  2, 15,  8,  1,
	13,  1,  6, 10,  4, 13,  9,  0,  8,  6, 15,  9,  3,  8,  0,  7,
	11,  4,  1, 15,  2, 14, 12,  3,  5, 11, 10,  5, 14,  2,  7, 12,
    };

    static final byte[] S4Table={
	 7, 13, 13,  8, 14, 11,  3,  5,  0,  6,  6, 15,  9,  0, 10,  3,
	 1,  4,  2,  7,  8,  2,  5, 12, 11,  1, 12, 10,  4, 14, 15,  9,
	10,  3,  6, 15,  9,  0,  0,  6, 12, 10, 11,  1,  7, 13, 13,  8,
	15,  9,  1,  4,  3,  5, 14, 11,  5, 12,  2,  7,  8,  2,  4, 14,
    };
    static final byte[] S5Table={
	 2, 14, 12, 11,  4,  2,  1, 12,  7,  4, 10,  7, 11, 13,  6,  1,
	 8,  5,  5,  0,  3, 15, 15, 10, 13,  3,  0,  9, 14,  8,  9,  6,
	 4, 11,  2,  8,  1, 12, 11,  7, 10,  1, 13, 14,  7,  2,  8, 13,
	15,  6,  9, 15, 12,  0,  5,  9,  6, 10,  3,  4,  0,  5, 14,  3,
    };
    static final byte[] S6Table={
	12, 10,  1, 15, 10,  4, 15,  2,  9,  7,  2, 12,  6,  9,  8,  5,
	 0,  6, 13,  1,  3, 13,  4, 14, 14,  0,  7, 11,  5,  3, 11,  8,
	 9,  4, 14,  3, 15,  2,  5, 12,  2,  9,  8,  5, 12, 15,  3, 10,
	 7, 11,  0, 14,  4,  1, 10,  7,  1,  6, 13,  0, 11,  8,  6, 13,
    };
    static final byte[] S7Table={
	 4, 13, 11,  0,  2, 11, 14,  7, 15,  4,  0,  9,  8,  1, 13, 10,
	 3, 14, 12,  3,  9,  5,  7, 12,  5,  2, 10, 15,  6,  8,  1,  6,
	 1,  6,  4, 11, 11, 13, 13,  8, 12,  1,  3,  4,  7, 10, 14,  7,
	10,  9, 15,  5,  6,  0,  8, 15,  0, 14,  5,  2,  9,  3,  2, 12,
    };
    static final byte[] S8Table={
	13,  1,  2, 15,  8, 13,  4,  8,  6, 10, 15,  3, 11,  7,  1,  4,
	10, 12,  9,  5,  3,  6, 14, 11,  5,  0,  0, 14, 12,  9,  7,  2,
	 7,  2, 11,  1,  4, 14,  1,  7,  9,  4, 12, 10, 14,  8,  2, 13,
	 0, 15,  6, 12, 10,  9, 13,  0, 15,  3,  3,  5,  5,  6,  8, 11,
    };
    
    // STables[i-1][s] is the output for input s to S-box i
    static final byte[][] STables={S1Table, S2Table, S3Table, S4Table, S5Table, S6Table, S7Table, S8Table};

    static long SBox(long S) // input is a 48-bit integer stored in 64-bit signed "long"
    {
		// Split I into eight 6-bit chunks
		int Sa=(int)((S>>42));
		int Sb=(int)((S>>36)&63);
		int Sc=(int)((S>>30)&63);
		int Sd=(int)((S>>24)&63);
		int Se=(int)((S>>18)&63);
		int Sf=(int)((S>>12)&63);
		int Sg=(int)((S>>6)&63);
		int Sh=(int)(S&63);
		// Apply the S-boxes
		byte Oa=S1Table[Sa];
		byte Ob=S2Table[Sb];
		byte Oc=S3Table[Sc];
		byte Od=S4Table[Sd];
		byte Oe=S5Table[Se];
		byte Of=S6Table[Sf];
		byte Og=S7Table[Sg];
		byte Oh=S8Table[Sh];
		// Combine answers into 32-bit output stored in 64-bit signed "long"
		long O=(long)Oa<<28 | (long)Ob<<24 | (long)Oc<<20 | (long)Od<<16 | (long)Oe<<12 | (long)Of<<8 | (long)Og<<4 | (long)Oh;
		return(O);
    }

    static long EBox(long R) // input is a 32-bit integer stored in 64-bit signed "long"
    {
		// compute each 6-bit component
		long Ea=(R>>27)&31 | (R&1)<<5;
		long Eb=(R>>23)&63;
		long Ec=(R>>19)&63;
		long Ed=(R>>15)&63;
		long Ee=(R>>11)&63;
		long Ef=(R>>7)&63;
		long Eg=(R>>3)&63;
		long Eh=(R>>31)&1 | (R&31)<<1;
		// 48-bit output stored in 64-bit signed "long"		
		long E=(long)Ea<<42 | (long)Eb<<36 | (long)Ec<<30 | (long)Ed<<24 | (long)Ee<<18 | (long)Ef<<12 | (long)Eg<<6 | (long)Eh;
		return(E);
    }

    static final int[] Pbits={
	16,  7, 20, 21,
	29, 12, 28, 17,
	 1, 15, 23, 26,
	 5, 18, 31, 10,
	 2,  8, 24, 14,
	32, 27,  3,  9,
	19, 13, 30,  6,
	22, 11,  4, 25
    };

    // this would have been a lot faster as fixed binary operations rather than a loop
    static long PBox(long O) // input is a 32-bit integer stored in 64-bit signed "long"
    {
		long P=0L;
		for(int i=0; i<32; i++)
		{
		    P|=((O>>(32-Pbits[i]))&1) << (31-i);
		}
		return(P);
    }

    // a helper method to call the external program "desencrypt" in the current directory
    // the parameter is the 64-bit plaintext to encrypt, returns the ciphertext
    static long callEncrypt(long P) throws IOException
    {
		Process process = Runtime.getRuntime().exec("./desencrypt "+Long.toHexString(P));
		String CString = (new BufferedReader(new InputStreamReader(process.getInputStream()))).readLine();
		
		// we have to go via BigInteger otherwise the signed longs cause incorrect parsing
		long C=new BigInteger(CString, 16).longValue(); 

		return(C);
    }

    //Random Plaintext
    static long randomPlaintext()
    {
    	Random prng=new Random();
    	long P=prng.nextLong();
    	return P;
    }

    static HashSet<Byte> iosolateK(int indexK)		//indexK is from 1 to 8
    {
     	HashSet<Byte> isolatedK = new HashSet <Byte> ();

     	for(int i=0;i<64;i++)
     	{
     		byte addnum=(byte)(i&0x3f);
     		isolatedK.add(addnum);
     	}

     	int[] PI = new int[64];

     	//Calculate PI: 
     	int j=0;
     	for(int i=0;i<64;i++)	//for each number, the sbox of it (i) xor with sbox of the deltaed number, if it match
     	{
     		int tempindex=(int)(sboxinputDifferentials[indexK-1]^((byte)(i&0xff)));
     		int differentialoutput=(int)(STables[indexK-1][tempindex]^STables[indexK-1][i]);
     		if(differentialoutput == sboxoutputDifferentials[indexK-1])
     		{
     			PI[j]=i;
     			j++;
     		}
     	}

     	//Debug/////////////////////////////////////////////////////////

/*     	for(int i=0;i<j;i++)
     	{
     		System.out.print(String.format("%02x ",PI[i]));
     	}
     	System.out.println();*/

     	////////////////////////////////////////////////////////////

/*     	for(int i=0;i<14;i++)
     	{
     		System.out.print(String.format("%02x ",PI[i]));
     	}
     	System.out.println();*/

     	////////////////////////////////////////////////////////////////

     	int trycount=0;
     	int intersect_flag=0;

     	while(trycount<100 && isolatedK.size()>2)
     	{
	     	long P=randomPlaintext();
	     	long P_delta=P^inputDifferentials[indexK-1];

	     	//Compute Ea
	     	long E=EBox(P&0x00000000ffffffffL);
    		long E_indexK=(E>>(6*(7-(indexK-1))))&0x3f;
			byte E_indexK_Byte=(byte)E_indexK;

	     	long cipherP=0;
	     	long cipherDeltaP=0;

	     	try
	     	{
				cipherP=callEncrypt(P);
	     		cipherDeltaP=callEncrypt(P_delta);
	     	}
	     	catch (IOException e)
	     	{
	     		System.out.println("Error");
	     	}
	     	
	     	long deltaCipherP=cipherP^cipherDeltaP;
	     	if(deltaCipherP!=outputDifferentials[indexK-1])
	     	{
	     		//System.out.println("Skip this");
	     		continue;
	     	}
	     	else
	     	{
	     		//Failed Approach
	     		//////////////////////////////////////////////////////////////////////////////////
	     		//isolated K once, compute E_indexK, and compute K_indexK based on E and PI
/*	     		if(intersect_flag==0)
	     		{
	     			intersect_flag=1;
	     			for(int i=0;i<14;i++)
	     			{
	     				byte addnum=(byte)((PI[i]^E_indexK_Byte)&0x3f);
	     				isolatedK.add(addnum);
	     			}
	     			if(isolatedK.isEmpty()==true)
			 		{
			 			System.out.println("Empty!!!");
			 		}
			 		else
			 		{
			 			showHashset(isolatedK);
			 		}
	     		}
	     		else
	     		{
	     			for(int i=0;i<14;i++)
	     			{
	     				byte addnum=(byte)((PI[i]^E_indexK_Byte)&0x3f);
	     				isolatedK_temp.add(addnum);
	     			}
	     			isolatedK.retainAll(isolatedK_temp);
	     			if(isolatedK_temp.isEmpty()==true)
				 	{
				 		System.out.println("Empty!!!");
				 	}
				 	else
				 	{
				 		showHashset(isolatedK_temp);
				 		showHashset(isolatedK);
				 	}
				 	System.exit(0);
				 	if(isolatedK.size()==2)
				 	{
				 		break;
				 	}
	     		}*/
	     		/////////////////////////////////////////////////////////////////////////////////////

	     		HashSet<Byte> isolatedK_temp = new HashSet<Byte> ();

	     		for(int i=0;i<64;i++)
	     		{
	     			byte tempresult=(byte)((i^E_indexK_Byte)&0x3f);
	     			for(int k=0;k<possibilityTable[indexK-1];k++)
	     			{
	     				if(tempresult==PI[k])
	     				{
	     					byte addnum=(byte)(i&0x3f);
	     					isolatedK_temp.add(addnum);
	     				}
	     			}
	     		}

	     		isolatedK.retainAll(isolatedK_temp);
	     		showHashset(isolatedK);
	     	}
	     	trycount++;
     	}
     	showHashset(isolatedK);
     	return isolatedK;
    }

    static void showHashset (HashSet<Byte> hs)
    {
    	for (Byte b:hs)
    	{
    		System.out.print(String.format("%02x ",b));
    	}
    	System.out.println();
    }
}